package YAXML::Reverse;

#
#    YAML to XML converter module
#    Copyright (C) 2008 Antonio García Domínguez
#
#    This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use 5.008008;
use strict;
use warnings;

require Exporter;
our @ISA       = qw(Exporter);
our @EXPORT_OK = qw(Convert ConvertFile LoadFile VERSION);
our $VERSION   = '0.3.3';

use strict;
use warnings;
use open IN  => ':encoding(UTF-8)';
use open OUT => ':utf8';

use XML::Writer;
use YAML::XS;
use Encode qw(decode);

use constant {
    YAML_TAG => "yaml:yaml",
    ELEMENT_TAG => "yaml:_",
    ANCHOR_ATTR => "yaml:anchor",
    ALIAS_ATTR  => "yaml:alias",
    KEY_TAG => "yaml:_key",
    VALUE_TAG => "yaml:_value",
    XMLNS_YAML_ATTR => "xmlns:yaml",
    XMLNS_YAML => "tag:yaml.org,2002",
};

################ Prototypes ###################

=head1 NAME

YAXML::Reverse - Perl module which converts YAML to XML

=head1 SYNOPSIS

  use YAXML::Reverse qw(Convert ConvertFile);
  my $xmlText  = ConvertFile("foo.yaml");
  my $yamlText = <<<EOF
- What's
- up
- doc?
EOF;
  my $xmlText2 = Convert($yamlText);

=head1 DESCRIPTION

YAML is a data serialization language which aims to be much lighter
and easier to the eyes, and still understandable by any machine. It is
indeed far better suited to the task than XML, which suffers from
trying to do everything and not doing anything very well, but XML has
been around for a while and there's a quite large number of
technologies which have no parallel in YAML, such as XML Schema,
RELAX-NG or XSLT.

Though there exists an experimental XML binding for YAML, called
YAXML, no such thing seems to exist in the other direction. This
module attempts to be that mapping. Additionally, this module's share
directory includes a revised version of the XSLT stylesheet offered at
the YAML site, which adds multi-document and _key/_value support to
it.

=head2 EXPORT

This module exports the Convert and ConvertFile functions which
convert a string or the contents of the file in the specified path to
a XML string, respectively.

=head1 SEE ALSO

- YAML::Syck is very good at loading YAML 1.0 input data.
- YAML::XS is what we use now, as it can load YAML 1.1 data (such as
  Firefox 3 bookmarks).
- XML::Writer is useful when writing XML documents, as it provides some
  additional error reporting functionalities of its own.

This module was originally written to be part of a XMLEye YAML
document type, see https://forja.rediris.es/projects/csl2-xmleye/.

=head1 AUTHOR

Antonio GarcE<iacute>a DomE<iacute>nguez, E<lt>nyoescape@gmail.comE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2008 by Antonio GarcE<iacute>a DomE<iacute>nguez

This library is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License, version 3 or
higher.

=head1 FUNCTIONS

All functions in this module are listed below.

=head2 Convert

  Usage:      converts a YAML input string into its corresponding XML string.
  Type:       public function.
  Parameters: takes a single scalar with the YAML input.
  Returns:    single scalar with the resulting XML text.
  Exceptions: those from the YAML parsing module.

=cut

sub Convert($);

=head2 ConvertFile

  Usage:      converts a YAML input file into its corresponding XML string.
  Type:       public function.
  Parameters: takes a single scalar with the path to the YAML input file.
  Returns:    single scalar with the resulting XML text.
  Exceptions: those from the YAML parsing module.

=cut

sub ConvertFile($);

=head2 LoadFile

  Usage:      loads a YAML input file into a Perl data structure.
  Type:       public function.
  Parameters: takes a single scalar with the path to the YAML input file.
  Returns:    Perl data structure resulting from parsing the YAML input file.
  Exceptions: those from the YAML parsing module.

=cut

sub LoadFile($);

# Private functions

sub _dumpXML($$$);
sub _generateXMLDocument(@);
sub _isValidXMLName($);
sub _scanLinks($);
sub _scanLinksAux($$$);
sub _startTag($$$$$%);

################ Funciones ####################

sub Convert($)
{
    my ($text) = @_;
    my (@docs)  = Load($text);
    return decode('UTF-8', _generateXMLDocument(@docs));
}

sub ConvertFile($)
{
    my ($path) = @_;
    my (@docs) = LoadFile($path);
    return decode('UTF-8', _generateXMLDocument(@docs));
}

sub LoadFile($)
{
    my ($path) = @_;

    local *IF;
    open(IF, "<$path");
    my @lines   = <IF>;
    my $content = join '', @lines;
    my $data    = Load($content);
    
    return $data;
}

sub _generateXMLDocument(@)
{
    my (@data) = @_;

    # Set up the XML document
    my $xmlText;
    my $xml = new XML::Writer(OUTPUT => \$xmlText, ENCODING => "utf-8");
    $xml->xmlDecl();
    $xml->startTag(YAML_TAG, XMLNS_YAML_ATTR() => XMLNS_YAML);

    foreach my $doc (@data) {
	my $rhLinks = _scanLinks($doc);

	# Print the XML document header (unless it has a tag)
	unless (ref($doc) ne "ARRAY" && ref($doc) ne "HASH") {
	    $xml->startTag(YAML_TAG);
	    _dumpXML($xml, $doc, $rhLinks);
	    $xml->endTag(YAML_TAG);
	}
	else {
	    _dumpXML($xml, $doc, $rhLinks);
	}
    }

    # Close the root element and mark the document as done
    $xml->endTag(YAML_TAG);
    $xml->end();

    return $xmlText;
}

sub _scanLinks($) {
    my ($doc)     = @_;
    my $rhLinks = {};
    my $counter  = 0;
    _scanLinksAux($doc, $rhLinks, \$counter);

    foreach (keys %$rhLinks) {
	delete $rhLinks->{$_} 
  	   if $rhLinks->{$_} == 0;
    }
    return $rhLinks;
}

sub _scanLinksAux($$$) {
    my ($doc, $rhLinks, $rsCounter) = @_;

    if (ref $doc eq "ARRAY") {
	foreach (@$doc) {
	    next unless ref $_;

	    if (exists $rhLinks->{$_}) {
		$rhLinks->{$_} = ++$$rsCounter
		    unless $rhLinks->{$_} != 0;
	    }
	    else {
		# Scan recursively unless this is a link
		_scanLinksAux($_, $rhLinks, $rsCounter);
		$rhLinks->{$_} = 0;
	    }
	}
    }
    elsif (ref $doc) {
	foreach (values %$doc) {
	    next unless ref $_;

	    if (exists $rhLinks->{$_}) {
		$rhLinks->{$_} = ++$$rsCounter
		    unless $rhLinks->{$_} != 0;
	    }
	    else {
		# Scan recursively unless this is a link
		_scanLinksAux($_, $rhLinks, $rsCounter);
		$rhLinks->{$_} = 0;
	    }
	}
    }
}

sub _dumpXML($$$)
{
    my ($xml, $data, $rhLinks) = @_;

    my $tipoRef = ref $data;

    if ($tipoRef eq "ARRAY") {
	# List
	foreach my $elem (@$data) {
	    _startTag($xml, $elem, $rhLinks, 0, ELEMENT_TAG);
	}
    }
    elsif ($tipoRef) {
	my $nombre;
	if ($tipoRef ne "HASH") {
	    # A tagged document is modeled as an object, which is really a hash
	    (my $url, $nombre) = split m/::/, $tipoRef;
	    $nombre = YAML_TAG unless $nombre;
	    _startTag($xml, $data, $rhLinks, 1, $nombre, xmlns=>$url);
	}

	# Mapping
	while (my ($key, $val) = each(%$data)) {
	    if (_isValidXMLName($key)) {
		_startTag($xml, $val, $rhLinks, 0, $key);
	    } else {
		$xml->startTag(KEY_TAG);
		$xml->characters($key);
		$xml->endTag(KEY_TAG);
		_startTag($xml, $val, $rhLinks, 0, VALUE_TAG);
	    }
	}
	
	if ($tipoRef ne "HASH") {
	    $xml->endTag($nombre);
	}
    }
    else {
	# Scalar
	$xml->characters($data);
    }
    
}

sub _startTag($$$$$%)
{
    my ($xml, $elem, $rhLinks, $onlyInitialTag, $tag, %attrs) = @_;

    unless (exists $rhLinks->{$elem}) {
	if (ref $elem eq '') {
	    $attrs{"yaml:type"} = "scalar";
	}
	else {
	    $attrs{"yaml:type"} = lc ref $elem;
	}

	$xml->startTag($tag, %attrs);
	unless ($onlyInitialTag) {
	    _dumpXML($xml, $elem, $rhLinks);
	    $xml->endTag($tag);
	}
    }
    elsif ($rhLinks->{$elem} > 0) {
	# Shared reference in the link hadn't appeared yet, so it's an anchor
	$xml->startTag($tag, ANCHOR_ATTR() => "id" . $rhLinks->{$elem}, %attrs);
	unless ($onlyInitialTag) {
	    _dumpXML($xml, $elem, $rhLinks);
	    $xml->endTag($tag);
	}

	# Mark as seen
	$rhLinks->{$elem} = -$rhLinks->{$elem};
    }
    elsif ($rhLinks->{$elem} < 0) {
	# Shared reference has been seen already, so it's an alias
	$xml->emptyTag($tag, ALIAS_ATTR() => "id" . -$rhLinks->{$elem});
    }
}

sub _isValidXMLName($) 
{
    return 
	$_[0] !~ /:{2,}/ && 
	$_[0] =~ /^[A-Za-z_:][\w\.\-:]*$/;
}

1;

__END__
